﻿/*
 *         Progetto: Libreria di algoritmi di steganografia
 * 
 *           Autore: Roberto Fuligni
 *  Ultima modifica: 14/01/2019
 *  
 *      Descrizione: Funzioni di utilità impiegate all'interno della libreria.
 */

using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Runtime.InteropServices;
using System.Security.Cryptography;

namespace LibSteganografia
{
    class BitmapHelper : IDisposable
    {
        const int CountSize = 4;
        const int HashSize = 5;

        String _path = null;
        Bitmap _bmp = null;
        int _depth;
        IntPtr _Iptr = IntPtr.Zero;
        BitmapData _data = null;
        byte[] _pixels = null;
        bool _dirty = false;

        public int MaxContentLength { get; private set; }
        public BitmapHelper(string path, FileAccess access)
        {
            int pixelCount;
            _path = path;
            using (var fs = new FileStream(path, FileMode.Open, FileAccess.Read))
                _bmp = new Bitmap(fs);

            _depth = Bitmap.GetPixelFormatSize(_bmp.PixelFormat);

            if (_depth != 24 && _depth != 32)
                throw new ArgumentException($"La profondità di colore di {_depth} bit non è supportata");

            int step = _depth / 8;
            pixelCount = _bmp.Width * _bmp.Height * step;

            MaxContentLength = pixelCount / 8 - (CountSize + HashSize);    // 4 dimensione + 4 hash
            if (MaxContentLength < 0)
                throw new ArgumentException("Spazio insufficiente all'interno del file");

            Rectangle rect = new Rectangle(0, 0, _bmp.Width, _bmp.Height);
            _data = _bmp.LockBits(rect, ImageLockMode.ReadWrite, _bmp.PixelFormat);

            _pixels = new byte[pixelCount];
            _Iptr = _data.Scan0;

            Marshal.Copy(_Iptr, _pixels, 0, _pixels.Length);
        }

        // Inserimento di una sequenza di byte nei pixel di un'immagine bitmap.
        // Si memorizzzano, nell'ordine indicato, le seguenti informazioni:
        //     1. Lunghezza in byte della sequenza (4 byte)
        //     2. Hash MD5 della sequenza (16 byte)
        //     3. Contenuto della sequenza
        public void putContent(byte[] buffer)
        {
            if (buffer.Length > MaxContentLength)
                throw new IOException("Spazio insufficiente all'interno dell'immagine");


            byte[] countArray, hashArray, rawBuffer;
            MD5 h = MD5.Create();

            countArray = BitConverter.GetBytes(buffer.Length);
            hashArray = h.ComputeHash(countArray);
            rawBuffer = new byte[CountSize + HashSize + buffer.Length];
            Array.Copy(countArray, 0, rawBuffer, 0, CountSize);
            Array.Copy(hashArray, 0, rawBuffer, CountSize, HashSize);
            Array.Copy(buffer, 0, rawBuffer, CountSize + HashSize, buffer.Length);

            rawPutContent(rawBuffer);
        }

        // Routine di inserimento dei bit di informazione nell'immagine
        private void rawPutContent(byte[] rawBuffer, int skipBytes = 0)
        {
            const byte mask = Byte.MaxValue - 1;
            int step = _depth / 8;
            int skipBits = skipBytes * 8;
            int idx = (skipBits / 3) * step;
            skipBits /= 3;
            idx += (_depth == 32 ? skipBits + 1 : skipBits);

            for (int r = 0; r < rawBuffer.Length; r++)
            {
                byte b = rawBuffer[r];

                for (int ib = 0; ib < 8; ib++)
                {
                    if (_depth == 32 && (idx % 4) == 0)
                        idx++;
                    int temp = _pixels[idx] & mask;
                    temp |= (b & 1);
                    _pixels[idx++] = (byte)temp;
                    b >>= 1;
                }
            }
            _dirty = true;
        }

        // Estrazione di una sequenza di byte dai pixel dell'immagine
        public byte[] getContent()
        {
            byte[] header = rawGetContent(0, CountSize + HashSize);
            byte[] countArray = rawGetContent(0, CountSize);
            byte[] hashArray = rawGetContent(CountSize, HashSize);
            byte[] hashArray2;
            MD5 h = MD5.Create();


            hashArray2 = h.ComputeHash(countArray);

            for (int i = 0; i < HashSize; i++)
            {
                if (hashArray[i] != hashArray2[i])
                    throw new IOException("Rappresentazione interna del contenuto non valida.");
            }

            int cnt = BitConverter.ToInt32(countArray, 0);

            return rawGetContent(CountSize + HashSize, cnt);

        }

        // Routine di estrazione dei bit di informazione nell'immagine
        private byte[] rawGetContent(int skipBytes, int count)
        {
            const byte mask = (Byte.MaxValue / 2) + 1;
            int step = _depth / 8;
            int skipBits = skipBytes * 8;
            int idx = (skipBits / 3) * step;
            skipBits %= 3;
            idx += (_depth == 32 ? skipBits + 1 : skipBits);

            byte[] result = new byte[count];
            for (int i = 0; i < count; i++)
            {
                byte b = 0;
                for (int ib = 0; ib < 8; ib++)
                {
                    b >>= 1;
                    if (_depth == 32 && (idx % 4) == 0)
                        idx++;
                    if ((_pixels[idx++] & 1) != 0)
                        b |= mask;
                }
                result[i] = b;
            }

            return result;
        }

        public void Dispose()
        {
            if (_dirty)
            {
                Marshal.Copy(_pixels, 0, _Iptr, _pixels.Length);
                _bmp.UnlockBits(_data);
                _bmp.Save(_path);
                _dirty = false;
            }

        }
    }
}
